#include "wanpipe_fr.h"


/* SNMP
 ******************************************************************************
 *			fr_snmp_data()	
 *
 * Description: Save snmp request and parameters in private structure, enable
 *		TIMER interrupt, put current process in sleep.
 * Arguments:   
 * Returns:
 ******************************************************************************
 */
#define   FRDLCMIIFINDEX        3
#define   FRDLCMISTATE          4
#define   FRDLCMIADDRESS        5
#define   FRDLCMIADDRESSLEN     6
#define   FRDLCMIPOLLINGINTERVAL  7
#define   FRDLCMIFULLENQUIRYINTERVAL  8
#define   FRDLCMIERRORTHRESHOLD  9
#define   FRDLCMIMONITOREDEVENTS  10
#define   FRDLCMIMAXSUPPORTEDVCS  11
#define   FRDLCMIMULTICAST      12
#define   FRDLCMISTATUS         13
#define   FRDLCMIROWSTATUS      14
#define   FRCIRCUITIFINDEX      17
#define   FRCIRCUITDLCI         18
#define   FRCIRCUITSTATE        19
#define   FRCIRCUITRECEIVEDFECNS  20
#define   FRCIRCUITRECEIVEDBECNS  21
#define   FRCIRCUITSENTFRAMES   22
#define   FRCIRCUITSENTOCTETS   23
#define   FRCIRCUITRECEIVEDFRAMES  24
#define   FRCIRCUITRECEIVEDOCTETS  25
#define   FRCIRCUITCREATIONTIME  26
#define   FRCIRCUITLASTTIMECHANGE  27
#define   FRCIRCUITCOMMITTEDBURST  28
#define   FRCIRCUITEXCESSBURST  29
#define   FRCIRCUITTHROUGHPUT   30
#define   FRCIRCUITMULTICAST    31
#define   FRCIRCUITTYPE         32
#define   FRCIRCUITDISCARDS     33
#define   FRCIRCUITRECEIVEDDES  34
#define   FRCIRCUITSENTDES      35
#define   FRCIRCUITLOGICALIFINDEX  36
#define   FRCIRCUITROWSTATUS    37
#define   FRERRIFINDEX          40
#define   FRERRTYPE             41
#define   FRERRDATA             42
#define   FRERRTIME             43
#define   FRERRFAULTS           44
#define   FRERRFAULTTIME        45
#define   FRTRAPSTATE           46
#define   FRTRAPMAXRATE         47

int wp_fr_snmp(void* chan_ptr, void* data)
{
	fr_channel_t* 	chan 	= (fr_channel_t*)chan_ptr;
	fr_prot_t*	fr_prot	= chan->prot;
	wanpipe_snmp_t*	snmp 	= (wanpipe_snmp_t*)data;
	
	switch(snmp->snmp_magic){

	/***** Data Link Connection Management Interface ****/
	case FRDLCMIIFINDEX:
		break;

	case FRDLCMISTATE:
		 snmp->snmp_val = 
		 (fr_prot->cfg.signalling == WANOPT_FR_ANSI) ? SNMP_FR_ANSIT1617D :
		 (fr_prot->cfg.signalling == WANOPT_FR_Q933) ? SNMP_FR_ITUT933A :
		 (fr_prot->cfg.signalling == WANOPT_FR_LMI) ? SNMP_FR_LMIREV:
		       						SNMP_FR_NOLMICONF;
		break;

	case FRDLCMIADDRESS:
		snmp->snmp_val = SNMP_FR_Q922;
		break;

	case FRDLCMIADDRESSLEN:
		snmp->snmp_val = SNMP_FR_4BYTE_ADDR;
		break;

	case FRDLCMIPOLLINGINTERVAL:
		snmp->snmp_val = fr_prot->cfg.t391;
		break;

	case FRDLCMIFULLENQUIRYINTERVAL:
		snmp->snmp_val = fr_prot->cfg.n391;
		break;

	case FRDLCMIERRORTHRESHOLD:
		snmp->snmp_val = fr_prot->cfg.n392;
		break;

	case FRDLCMIMONITOREDEVENTS:
		snmp->snmp_val = fr_prot->cfg.n393;
		break;

	case FRDLCMIMAXSUPPORTEDVCS:
		snmp->snmp_val = HIGHEST_VALID_DLCI;
		break;

	case FRDLCMIMULTICAST:
		snmp->snmp_val = SNMP_FR_NONBROADCAST;
		break;

	case FRDLCMISTATUS: /* FIXME */
		snmp->snmp_val = SNMP_FR_RUNNING;
		break;

	case FRDLCMIROWSTATUS: /* FIXME */
		snmp->snmp_val = 0;
		break;

	/****************** Circuit Table *******************/
	case FRCIRCUITIFINDEX: 
		break;

	case FRCIRCUITDLCI:
		snmp->snmp_val = chan->dlci;
		break;

	case FRCIRCUITSTATE:
		snmp->snmp_val = (chan->dlci_state & PVC_STATE_ACTIVE) ? SNMP_FR_ACTIVE : SNMP_FR_INACTIVE;
		break;

	case FRCIRCUITRECEIVEDFECNS:
		snmp->snmp_val = chan->rx_FECN;
		break;

	case FRCIRCUITRECEIVEDBECNS:
		snmp->snmp_val = chan->rx_BECN;
		break;

	case FRCIRCUITSENTFRAMES:
		snmp->snmp_val = chan->dlci_stats.tx_frames;
		break;

	case FRCIRCUITSENTOCTETS:
		snmp->snmp_val = chan->dlci_stats.tx_bytes;
		break;

	case FRCIRCUITRECEIVEDFRAMES:
		snmp->snmp_val = chan->dlci_stats.rx_frames;
		break;

	case FRCIRCUITRECEIVEDOCTETS:
		snmp->snmp_val = chan->dlci_stats.rx_bytes;
		break;

	case FRCIRCUITCREATIONTIME:
#if 0
		wapbs_do_gettimeofday(&tv);
		snmp->snmp_val = tv.tv_sec - chan->router_start_time;
#else
		snmp->snmp_val = 0;
#endif
		break;

	case FRCIRCUITLASTTIMECHANGE:
#if 0
		wpabs_do_gettimeofday(&tv);
		snmp->snmp_val = tv.tv_sec - chan->router_last_change;
#else
		snmp->snmp_val = 0;
#endif
		break;

	case FRCIRCUITCOMMITTEDBURST:
		snmp->snmp_val = (unsigned long)0;
		break;

	case FRCIRCUITEXCESSBURST:
		snmp->snmp_val = (unsigned long)0;
		break;

	case FRCIRCUITTHROUGHPUT: /* FIXME */
		snmp->snmp_val = (unsigned long)0;
		break;

	case FRCIRCUITMULTICAST:
		/* FIXME */
#if 0
		snmp->snmp_val = (chan->mc) ? SNMP_FR_ONEWAY : SNMP_FR_UNICAST;
#else
		snmp->snmp_val = SNMP_FR_UNICAST;
#endif
		break;

	case FRCIRCUITTYPE:
		snmp->snmp_val = SNMP_FR_STATIC;
		break;

	case FRCIRCUITDISCARDS:
		snmp->snmp_val = fr_prot->link_stats.rx_bad_format;
		break;

	case FRCIRCUITRECEIVEDDES:
		snmp->snmp_val = 0; /* FIXME chan->rx_DE_set; */
		break;

	case FRCIRCUITSENTDES:
		snmp->snmp_val = 0; /* FIXME chan->tx_DE_set; */
		break;

	case FRCIRCUITLOGICALIFINDEX: /* FIXME */
		snmp->snmp_val = 0;
		break;

	case FRCIRCUITROWSTATUS: /* FIXME */
		snmp->snmp_val = 0;
		break;

	/****************** Errort Table  *******************/
	case FRERRIFINDEX: 
		break;

	case FRERRTYPE: 
		snmp->snmp_val = 0; /*FIXME chan->err_type;*/
		break;

	case FRERRDATA: /* FIXME */
#if 0
		strcpy((void*)snmp->snmp_data, chan->err_data);
#endif
		break;

	case FRERRTIME: /* FIXME */
#if 0
		snmp->snmp_val = chan->err_time;
#else
		snmp->snmp_val = 0;
#endif
		break;

	case FRERRFAULTS: /* FIXME */
#if 0
		snmp->snmp_val = chan->err_faults;
#else
		snmp->snmp_val = 0;
#endif
		break;

	case FRERRFAULTTIME: /* FIXME */
		snmp->snmp_val = (unsigned long)0;
		break;

	/************ Frame Relay Trap Control **************/
	case FRTRAPSTATE:
		snmp->snmp_val = SNMP_FR_DISABLED; /* FIXME chan->trap_state; */
		break;

	case FRTRAPMAXRATE:
		snmp->snmp_val = 0; /*FIXME chan->trap_max_rate;*/
		break;

	default:
            	return -1;
	}

	return 0;
}

