/*
 *	LAPB release 002
 *
 *	This code REQUIRES 2.1.15 or higher/ NET3.038
 *
 *	This module:
 *		This module is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 */
#include "wanpipe_sppp.h"

static void wpsppp_lcp_input (struct sppp *sp, void *m);
static void wpsppp_cisco_input (struct sppp *sp, void *m);
static void wpsppp_ipcp_input (struct sppp *sp, void *m);


static void wpsppp_ipcp_input (struct sppp *sp, void *skb)
{
	struct lcp_header *h;
	struct ipcp_header *ipcp_h;
	int len = wpabs_skb_len(skb);
	unsigned int remote_ip, local_ip;
	

	if (len < 4) 
	{
		sp->prot_stats.rx_unknown++;
		if (sp->pp_flags & PP_DEBUG) {
			wpabs_debug_event( "%s: invalid ipcp packet length: %d bytes\n",
				sp->name,  len);
		}
		return;
	}
	
	
	h = (struct lcp_header *)wpabs_skb_data(skb);
	wpabs_skb_pull(skb,sizeof(struct lcp_header));
	if (sp->pp_flags & PP_DEBUG) {
		char state = '?';
		switch (sp->ipcp.state) {
		case IPCP_STATE_CLOSED:   state = 'C'; break;
		case IPCP_STATE_ACK_RCVD: state = 'R'; break;
		case IPCP_STATE_ACK_SENT: state = 'S'; break;
		case IPCP_STATE_OPENED:   state = 'O'; break;
		}
		
		wpabs_debug_event( "%s: ipcp input(%c): %d bytes <%s id=%xh len=%xh",
			sp->name,  
			state,
			len,
			wpsppp_ipcp_type_name (h->type), h->ident, wpabs_ntohs (h->len));

		if (len > 4)
			wpsppp_print_bytes ((unsigned char*) (h+1), len-4);

		wpabs_debug_event (">\n");
	}

	if (len > wpabs_ntohs (h->len))
		len = wpabs_ntohs (h->len);

	ipcp_h = (struct ipcp_header*)(h+1);

	switch (h->type) {
	default:
		/* Unknown packet type -- send Code-Reject packet. */
		sp->prot_stats.rx_unknown++;
		sp->prot_stats.tx_code_rej++;
		wpsppp_cp_send (sp, PPP_IPCP, IPCP_CODE_REJ, ++sp->pp_seq[IDX_LCP], len, h);
		break;

	case IPCP_CONF_REQ:

		if (len < 4) {
			sp->prot_stats.rx_unknown++;
			if (sp->pp_flags & PP_DEBUG) {
				wpabs_debug_event( "%s: invalid ipcp configure request packet length: %d bytes\n",
					sp->name, len);
			}
			return;
		}
		sp->prot_stats.rx_conf_rqst++;
		if (len > 4) {

			if (sp->dynamic_ip != WANOPT_PPP_PEER){

				remote_ip=wpsppp_get_ipv4addr(sp,WAN_POINTOPOINT_IP);
				
				if (remote_ip == *(unsigned int*)&ipcp_h->data[0] || 
				    wpsppp_get_ipv4addr(sp,WAN_LOCAL_IP) == *(unsigned int*)&ipcp_h->data[0]){
					sp->prot_stats.tx_conf_ack++;
					wpsppp_cp_send (sp, PPP_IPCP, LCP_CONF_ACK, h->ident,
					len-4, h+1);
                                         
					switch (sp->ipcp.state) {
					case IPCP_STATE_OPENED:
						break;
					case IPCP_STATE_ACK_RCVD:	
			       	 		sp->ipcp.state = IPCP_STATE_OPENED;
				        	DBG_CHDLC("%s(): line: %d: IPCP (IPver) CFG REQ: OPEN\n", 
				        		__FUNCTION__, __LINE__); 
						break;
					default:
						sp->ipcp.state = IPCP_STATE_ACK_SENT;
				        	DBG_CHDLC("%s(): line: %d: IPCP (IPver) CFG REQ: ACK SENT\n", 
				        		__FUNCTION__, __LINE__); 
						break;
					}              
					
					remote_ip=*(unsigned int*)&ipcp_h->data[0];
					wpabs_debug_event("%s: IPCP Static: P-to-P verified: %u.%u.%u.%u\n",
							sp->name,
							NIPQUAD(remote_ip));

				} else if (*(unsigned int*)&ipcp_h->data[0] == 0 ) {
				     	unsigned char data[]={3,6,0,0,0,0};      
				      	*(unsigned int*)&data[2]=wpsppp_get_ipv4addr(sp,WAN_POINTOPOINT_IP); 

                                        wpabs_debug_event("%s: IPCP Static: IP Req Received\n",
							sp->name);
					
                                       	 wpsppp_cp_send (sp, PPP_IPCP, IPCP_CONF_NAK, h->ident,
								6, data);   

					 /* Change the state. */
					switch (sp->ipcp.state) {
					case IPCP_STATE_OPENED:
						break;
					case IPCP_STATE_ACK_RCVD:	
			       	 		sp->ipcp.state = IPCP_STATE_OPENED;
				        	DBG_CHDLC("%s(): line: %d: IPCP (IPreq) CFG REQ: OPEN\n", 
				        		__FUNCTION__, __LINE__); 
						break;
					default:
						sp->ipcp.state = IPCP_STATE_ACK_SENT;
				        	DBG_CHDLC("%s(): line: %d: IPCP (IPreq) CFG REQ: ACK SENT\n", 
				        		__FUNCTION__, __LINE__); 
						break;
					}

				} else {

					remote_ip = *(unsigned int*)&ipcp_h->data[0];

					sp->local_ip=wpsppp_get_ipv4addr(sp,WAN_LOCAL_IP);
					sp->remote_ip=wpsppp_get_ipv4addr(sp,WAN_POINTOPOINT_IP);	
					wpabs_debug_event("%s: IPCP Static Refusing P-to-P %u.%u.%u.%u: Dynamic ip disabled!\n",
							sp->name, NIPQUAD(remote_ip));

					wpabs_debug_event("%s: IPCP Current Cfg: Local %u.%u.%u.%u P-to-P %u.%u.%u.%u\n",
							sp->name,
							NIPQUAD(sp->local_ip),
							NIPQUAD(sp->remote_ip));

					if (++sp->dynamic_failures > 10){
				
						sp->prot_stats.tx_term_rqst++;
						wpsppp_cp_send (sp, PPP_IPCP, IPCP_TERM_REQ, h->ident, 0, 0);
						sp->ipcp.state = IPCP_STATE_CLOSED;
						wpsppp_set_timeout (sp, 5);
						sp->dynamic_failures=0;
						
					}else{
					
						sp->prot_stats.tx_conf_rej++;
						wpsppp_cp_send (sp, PPP_IPCP, LCP_CONF_REJ, h->ident,
							len-4, h+1);

						switch (sp->ipcp.state) {
						case IPCP_STATE_OPENED:
							/* Initiate renegotiation. */
							wpsppp_ipcp_open (sp);
							/* fall through... */
						case IPCP_STATE_ACK_SENT:
							/* Go to closed state. */
							sp->ipcp.state = IPCP_STATE_CLOSED;
						}
					}
				}

			} else {
				sp->prot_stats.tx_conf_ack++;
				sp->remote_ip = *(unsigned int*)&ipcp_h->data[0];
				wpsppp_cp_send (sp, PPP_IPCP, LCP_CONF_ACK, h->ident,
				len-4, h+1);
				
				switch (sp->ipcp.state) {
				case IPCP_STATE_OPENED:
					break;
				case IPCP_STATE_ACK_RCVD:	
			       		sp->ipcp.state = IPCP_STATE_OPENED;
				       	DBG_CHDLC("%s(): line: %d: IPCP (DIP) CFG REQ: OPEN\n", 
				       		__FUNCTION__, __LINE__); 
					break;
				default:
					sp->ipcp.state = IPCP_STATE_ACK_SENT;
				       	DBG_CHDLC("%s(): line: %d: IPCP (DIP) CFG REQ: ACK SENT\n", 
				      		__FUNCTION__, __LINE__); 
					break;
				}              
			}
		} else {
			/* Send Configure-Ack packet. */
			sp->prot_stats.tx_conf_ack++;
			wpsppp_cp_send (sp, PPP_IPCP, IPCP_CONF_ACK, h->ident,
				0, 0);

			switch (sp->ipcp.state) {
			case IPCP_STATE_OPENED:
				break;
			case IPCP_STATE_ACK_RCVD:	
				sp->ipcp.state = IPCP_STATE_OPENED;
			       	DBG_CHDLC("%s(): line: %d: IPCP (NoIP) CFG REQ: OPEN\n", 
			       		__FUNCTION__, __LINE__); 
				break;
			default:
				sp->ipcp.state = IPCP_STATE_ACK_SENT;
			       	DBG_CHDLC("%s(): line: %d: IPCP (NoIP) CFG REQ: ACK SENT\n", 
			      		__FUNCTION__, __LINE__); 
				break;
			}       
		}
		break;
		
	case IPCP_CONF_ACK:
		if (h->ident != sp->ipcp.confid)
			break;
		
		sp->prot_stats.rx_conf_ack++;
		
		wpsppp_clear_timeout (sp);
		
		switch (sp->ipcp.state) {
		case IPCP_STATE_CLOSED:
			sp->ipcp.state = IPCP_STATE_ACK_RCVD;
			DBG_CHDLC("%s(): line: %d: IPCP ACK: ACK RCVD\n", 
						__FUNCTION__, __LINE__);     
			wpsppp_set_timeout (sp, 5);
			break;
		case IPCP_STATE_ACK_SENT:
			sp->ipcp.state = IPCP_STATE_OPENED;
			DBG_CHDLC("%s(): line: %d: IPCP ACK: OPENED\n", 
						__FUNCTION__, __LINE__);
			if (sp->dynamic_ip == WANOPT_PPP_PEER){
				sp->dynamic_event=1;
			}
			break;
		}
		break;

	case IPCP_CONF_NAK:

		sp->prot_stats.rx_conf_nak++;

		if (ipcp_h->len >= 6) {

			if (sp->dynamic_ip != WANOPT_PPP_PEER) {

				local_ip=wpsppp_get_ipv4addr(sp,WAN_LOCAL_IP);
	
				if (local_ip == *(unsigned int*)&ipcp_h->data[0]) {
					sp->lcp_stats.tx_conf_rqst++;
			       	 	wpsppp_cp_send (sp, PPP_IPCP, LCP_CONF_REQ, h->ident,
				       	 len-4, h+1);
					
			       		sp->ipcp.state = IPCP_STATE_ACK_SENT;  

					local_ip=*(unsigned int*)&ipcp_h->data[0];	
					wpabs_debug_event("%s: IPCP Static: Local IP verified: %u.%u.%u.%u\n",
							sp->name,
							NIPQUAD(local_ip));

					break;
				} else {
					local_ip=*(unsigned int*)&ipcp_h->data[0]; 
					wpabs_debug_event("%s: IPCP Static: Refusing Local %u.%u.%u.%u: Dynamic ip disabled!\n",
							sp->name, NIPQUAD(local_ip));

					wpabs_debug_event("%s: IPCP Current Cfg: Local %u.%u.%u.%u P-to-P %u.%u.%u.%u\n",
							sp->name,
							NIPQUAD(sp->local_ip),
							NIPQUAD(sp->remote_ip));
				
					sp->prot_stats.tx_conf_rej++;
					wpsppp_cp_send (sp, PPP_IPCP, LCP_CONF_REJ, h->ident,
						len-4, h+1);

					/* Drop down to reject */

				}

			} else {
				sp->local_ip = *(unsigned int*)&ipcp_h->data[0];
				sp->lcp_stats.tx_conf_rqst++;
				wpsppp_cp_send (sp, PPP_IPCP, LCP_CONF_REQ, h->ident,
					len-4, h+1);
					                 
				sp->ipcp.state = IPCP_STATE_ACK_SENT;
				break;
			}
		}
		/* Drop down to reject */
		
	case IPCP_CONF_REJ:
		
		if (h->ident != sp->ipcp.confid)
			break;
		
		sp->prot_stats.rx_conf_rej++;
		
		DBG_CHDLC("%s(): line: %d  IPCP REJ\n", __FUNCTION__, __LINE__);
		//wpsppp_clear_timeout (sp);
		//wpsppp_ipcp_open (sp);
		
		/* Initiate renegotiation. */
		if (sp->ipcp.state != IPCP_STATE_ACK_SENT) {
			/* Go to closed state. */
			sp->ipcp.state = IPCP_STATE_CLOSED;
		       	DBG_CHDLC("%s(): line: %d  IPCP Closed\n", __FUNCTION__, __LINE__);
			
		}

		wpsppp_clear_timeout (sp);
		wpsppp_set_timeout (sp,2);   
		break;

	case IPCP_TERM_REQ:
		
		sp->prot_stats.rx_term_rqst++;
		
		/* Send Terminate-Ack packet. */
		sp->prot_stats.tx_term_ack++;
		wpsppp_cp_send (sp, PPP_IPCP, IPCP_TERM_ACK, h->ident, 0, 0);
		/* Go to closed state. */
		sp->ipcp.state = IPCP_STATE_CLOSED;
		/* Initiate renegotiation. */
		wpsppp_ipcp_open (sp);
		break;

	case IPCP_TERM_ACK:
		sp->prot_stats.rx_term_ack++;
		/* Ignore for now. */
		break;
	case IPCP_CODE_REJ:
		sp->prot_stats.rx_code_rej++;
		/* Ignore for now. */
		break;
	}
}



/*
 * Handle incoming Cisco keepalive protocol packets.
 */

static void wpsppp_cisco_input (struct sppp *sp, void *skb)
{
	struct cisco_packet *h;


	if (wpabs_skb_len(skb) != CISCO_PACKET_LEN && wpabs_skb_len(skb) != CISCO_BIG_PACKET_LEN) {
		if (sp->pp_flags & PP_DEBUG) {
			wpabs_debug_event( "%s: invalid cisco packet length: %d bytes\n",
				sp->name,  wpabs_skb_len(skb));
		}
		sp->chdlc_stats.Rx_Data_discard_short_count++;
		return;
	}

	h = (struct cisco_packet *)wpabs_skb_data(skb);
	wpabs_skb_pull(skb, sizeof(struct cisco_packet*));
	
	if (sp->pp_flags & PP_DEBUG) {
		wpabs_debug_event( "%s: cisco input: %d bytes <%xh %xh %xh %xh %xh-%xh>\n",
			sp->name,  wpabs_skb_len(skb),
			wpabs_ntohl (h->type), h->par1, h->par2, h->rel,
			h->time0, h->time1);
	}
	
	switch (wpabs_ntohl (h->type)) {
	default:
		if (sp->pp_flags & PP_DEBUG){
			wpabs_debug_event( "%s: unknown cisco packet type: 0x%x\n",
				sp->name,  wpabs_ntohl (h->type));
		}
		sp->chdlc_stats.Rx_invalid_CHDLC_type_count++;
		break;

	case CISCO_ADDR_REPLY:
		sp->chdlc_stats.CHDLC_SLARP_REPLY_Rx_count++;
		/* Reply on address request, ignore */
		break;
		
	case CISCO_KEEPALIVE_REQ:
		sp->chdlc_stats.CHDLC_SLARP_KPALV_Rx_count++;
		
		sp->pp_alivecnt = 0;
		sp->pp_rseq[IDX_LCP] = wpabs_ntohl (h->par1);
		if (sp->pp_seq[IDX_LCP] == sp->pp_rseq[IDX_LCP]) {
			/* Local and remote sequence numbers are equal.
			 * Probably, the line is in loopback mode. */
			int newseq;
			if (sp->pp_loopcnt >= sp->sppp_max_keepalive_count) {
				wpabs_debug_event( "%s: loopback\n",
					sp->name);
				sp->pp_loopcnt = 0;
				sp->pp_link_state=SPPP_LINK_DOWN;
				wpsppp_disconnect_indication(sp, 0);

			}
			++sp->pp_loopcnt;

			/* Generate new local sequence number */
			wpabs_get_random_bytes(&newseq, sizeof(newseq));
			sp->pp_seq[IDX_LCP] ^= newseq;
			break;
		}
		sp->pp_loopcnt = 0;
		if (sp->pp_link_state==SPPP_LINK_DOWN) {
			sp->pp_link_state=SPPP_LINK_UP;
			wpabs_debug_event( "%s: protocol up\n", sp->name);
			wpsppp_connect_indication(sp, 0);
		}
		break;
	case CISCO_ADDR_REQ:
		sp->chdlc_stats.CHDLC_SLARP_REQ_Rx_count++;
		
		/* Stolen from net/ipv4/devinet.c -- SIOCGIFADDR ioctl */
		{
		unsigned int addr=wpsppp_get_ipv4addr(sp,WAN_LOCAL_IP);
		unsigned int mask=wpsppp_get_ipv4addr(sp,WAN_NETMASK_IP);

		/* I hope both addr and mask are in the net order */
		wpsppp_cisco_send (sp, CISCO_ADDR_REPLY, addr, mask);
		}
		break;
	}
}



/*
 * Handle incoming PPP Link Control Protocol packets.
 */
 
static void wpsppp_lcp_input (struct sppp *sp, void *skb)
{
	struct lcp_header *h;
	int len = wpabs_skb_len(skb);
	unsigned char *p, opt[6];
	unsigned int rmagic;
	int rc_from_lcp_options = 0;

	FUNC_BEGIN();

	if (len < 4) {
		if (sp->pp_flags & PP_DEBUG) {
			wpabs_debug_event( "%s: invalid lcp packet length: %d bytes\n",
				sp->name, len);
		}
		sp->lcp_stats.rx_too_large++;
		return;
	}
	h = (struct lcp_header *)wpabs_skb_data(skb);
	wpabs_skb_pull(skb,sizeof(struct lcp_header *));
	
	if (sp->pp_flags & PP_DEBUG) {

		char state = '?';
		switch (sp->lcp.state) {
		case LCP_STATE_CLOSED:   state = 'C'; break;
		case LCP_STATE_ACK_RCVD: state = 'R'; break;
		case LCP_STATE_ACK_SENT: state = 'S'; break;
		case LCP_STATE_OPENED:   state = 'O'; break;
		}
		wpabs_debug_event("%s: lcp input(%c): %d bytes <%s id=%xh len=%xh",
			sp->name, state, len,
			wpsppp_lcp_type_name (h->type), h->ident, wpabs_ntohs (h->len));
		if (len > 4)
			wpsppp_print_bytes ((unsigned char*) (h+1), len-4);
		wpabs_debug_event (">\n");
	}

	if (len > wpabs_ntohs (h->len))
		len = wpabs_ntohs (h->len);
	switch (h->type) 
	{
	default:
		/* Unknown packet type -- send Code-Reject packet. */
		sp->lcp_stats.tx_code_rej++;
		wpsppp_cp_send (sp, PPP_LCP, LCP_CODE_REJ, ++sp->pp_seq[IDX_LCP],
			wpabs_skb_len(skb), h);
		sp->lcp_stats.rx_unknown++;
		break;
		
	case LCP_CONF_REQ:

		if (len < 4) {
			if (sp->pp_flags & PP_DEBUG) {
				wpabs_debug_event("%s: invalid lcp configure request packet length: %d bytes\n",
					sp->name, len);
			}
			
			sp->lcp_stats.rx_too_large++;	
			break;
		}

		rc_from_lcp_options = wpsppp_lcp_conf_parse_options (sp, h, len, &rmagic);
		DBG_CHDLC("rc_from_lcp_options: 0x%X (%d)\n", rc_from_lcp_options, rc_from_lcp_options);

		if (len>4 && !rc_from_lcp_options){
			sp->lcp_stats.rx_too_large++;
			goto badreq;
		}

		sp->lcp_stats.rx_conf_rqst++;
		
		if (rmagic == sp->lcp.magic) {
			/* Local and remote magics equal -- loopback? */
			if (sp->pp_loopcnt >= sp->sppp_max_keepalive_count*5) {
				wpabs_debug_event( "%s: loopback\n",
					sp->name);
				sp->pp_loopcnt = 0;

				sp->pp_link_state=SPPP_LINK_DOWN;
				wpsppp_disconnect_indication(sp, 0);

			} else if (sp->pp_flags & PP_DEBUG) {
				wpabs_debug_event( "%s: conf req: magic glitch\n",
					sp->name);
			}
			
			++sp->pp_loopcnt;

			/* MUST send Conf-Nack packet. */
			rmagic = ~sp->lcp.magic;
			opt[0] = LCP_OPT_MAGIC;
			opt[1] = sizeof (opt);
			opt[2] = rmagic >> 24;
			opt[3] = rmagic >> 16;
			opt[4] = rmagic >> 8;
			opt[5] = rmagic;

			sp->lcp_stats.tx_conf_nak++;
			wpsppp_cp_send (sp, PPP_LCP, LCP_CONF_NAK,
				h->ident, sizeof (opt), &opt);
badreq:
			switch (sp->lcp.state) {
			case LCP_STATE_OPENED:
				/* Initiate renegotiation. */
				wpsppp_lcp_open (sp);
				/* fall through... */
			case LCP_STATE_ACK_SENT:
				/* Go to closed state. */
				sp->lcp.state = LCP_STATE_CLOSED;
				sp->ipcp.state = IPCP_STATE_CLOSED;
			}
			break;
		}/* if(rmgic...) */

		/* Send Configure-Ack packet. */
		sp->pp_loopcnt = 0;

		if (sp->lcp.state != LCP_STATE_OPENED) {
			sp->lcp_stats.tx_conf_ack++;
			wpsppp_cp_send (sp, PPP_LCP, LCP_CONF_ACK, h->ident, len-4, h+1);
		}

		/* 3/20/2006 Chris "Xenon" Hanson, xenon@arcticus.com
		We should not send PAP-REQ here, because we don't know if we have completed
		CFG-REQ and CFG-ACK in both directions yet (here we are completing only one side).
		Below, where we know we're in LCP_STATE_OPENED, we'll check to see if 
		wpsppp_lcp_conf_parse_options() set PP_NEEDAUTH in sp->pp_flags, and if so, then
		we'll transmit PAP-REQ. If we're doing PAP-REQ we won't go do wpsppp_ipcp_open(),
		because the code for PAP_ACK handling does that upon success.
		We also need to rig the code in LCP_CONF_ACK handling below to possibly transmit
		PAP-ACK if needed at that stage, because we don't know which order the two endpoints
		will execute the two LCP_CONF_REQ/LCP_CONF_ACK operations.
                */
#if 0	
		/* NC. Kernel change */
		wpsppp_cp_send (sp, PPP_LCP, LCP_CONF_ACK,
				h->ident, len-4, h+1);
#endif				

		/* Change the state. */
		switch (sp->lcp.state) 
		{
		case LCP_STATE_CLOSED:
			sp->lcp.state = LCP_STATE_ACK_SENT;
			break;

		case LCP_STATE_ACK_RCVD:
			/* 3/20/2006 CXH begin */
			if(sp->pp_flags & PP_NEEDAUTH){
				sp->myauth.authenticated = 0;
				if(rc_from_lcp_options == PPP_PAP){
					/* authenticator wants PAP. initiate PAP request. */
					sp->confid[IDX_PAP] = h->ident;
					wpsppp_pap_scr(sp);
					/* we don't want to continue to wpsppp_ipcp_open() 
				 	* yet, PAP_ACK will do it for us */
					break;
				}
			}else{
				sp->lcp.state = LCP_STATE_OPENED;
			}
			/* CXH end */

			wpsppp_ipcp_open (sp);
			break;

		case LCP_STATE_OPENED:
			/* Remote magic changed -- close session. */
			sp->lcp.state = LCP_STATE_CLOSED;
			sp->ipcp.state = IPCP_STATE_CLOSED;
			/* Initiate renegotiation. */
			wpsppp_lcp_open (sp);

			/* Send ACK after our REQ in attempt to break loop */
			sp->lcp_stats.tx_conf_ack++;
			wpsppp_cp_send (sp, PPP_LCP, LCP_CONF_ACK,
				h->ident, len-4, h+1);
			sp->lcp.state = LCP_STATE_ACK_SENT;
			break;
		}
		break;

	case LCP_CONF_ACK:
		
		if (h->ident != sp->lcp.confid){
			sp->lcp_stats.rx_unknown++;
			break;
		}
		
		sp->lcp_stats.rx_conf_ack++;
		
		wpsppp_clear_timeout (sp);
		
		if ((sp->pp_link_state != SPPP_LINK_UP) &&
		    !(sp->pp_flags & PP_NEEDAUTH)) {	/* David Rokhvarg: do not indicate 'up' until authentication finished */
			/* Coming out of loopback mode. */
			sp->pp_link_state=SPPP_LINK_UP;
			wpabs_debug_event( "%s: protocol up\n", sp->name);
			wpsppp_connect_indication(sp, 0);
		}
		
		switch (sp->lcp.state) {
		case LCP_STATE_CLOSED:
			sp->lcp.state = LCP_STATE_ACK_RCVD;
			wpsppp_set_timeout (sp, 5);
			break;
		case LCP_STATE_ACK_SENT:
			DBG_CHDLC("%s(): line: %d\n", __FUNCTION__, __LINE__);
			/* 3/20/2006 CXH begin */
			if(sp->pp_flags & PP_NEEDAUTH){ 
				DBG_CHDLC("%s(): line: %d\n", __FUNCTION__, __LINE__);
				if(sp->myauth.proto == PPP_PAP){
					/* authenticator wants PAP. initiate PAP request. */
					sp->confid[IDX_PAP] = h->ident;
					wpsppp_pap_scr(sp);
					/* we don't want to continue to wpsppp_ipcp_open() 
					 * yet, PAP_ACK will do it for us */
					break;  
				}
			}else{
				DBG_CHDLC("%s(): line: %d\n", __FUNCTION__, __LINE__);
				sp->lcp.state = LCP_STATE_OPENED;
			}
			/* CXH end */   
			
			wpsppp_ipcp_open (sp);
			break;
		}
		break;
	case LCP_CONF_NAK:
		if (h->ident != sp->lcp.confid){
			sp->lcp_stats.rx_unknown++;
			break;
		}
		
		sp->lcp_stats.rx_conf_nak++;
		
		p = (unsigned char*) (h+1);
		if (len>=10 && p[0] == LCP_OPT_MAGIC && p[1] >= 4) {
			rmagic = (unsigned int)p[2] << 24 |
				(unsigned int)p[3] << 16 | p[4] << 8 | p[5];
			if (rmagic == ~sp->lcp.magic) {
				int newmagic;
				if (sp->pp_flags & PP_DEBUG) {
					wpabs_debug_event( "%s: conf nak: magic glitch\n",
						sp->name);
				}
				wpabs_get_random_bytes(&newmagic, sizeof(newmagic));
				sp->lcp.magic += newmagic;
			} else
				sp->lcp.magic = rmagic;
			}
		if (sp->lcp.state != LCP_STATE_ACK_SENT) {
			/* Go to closed state. */
			sp->lcp.state = LCP_STATE_CLOSED;
			sp->ipcp.state = IPCP_STATE_CLOSED;
		}
		/* The link will be renegotiated after timeout,
		 * to avoid endless req-nack loop. */
		wpsppp_clear_timeout (sp);
		wpsppp_set_timeout (sp, 2);
		break;
		
	case LCP_CONF_REJ:
		if (h->ident != sp->lcp.confid){
			sp->lcp_stats.rx_unknown++;
			break;
		}
		
		sp->lcp_stats.rx_conf_rej++;

		if (sp->lcp.state != LCP_STATE_ACK_SENT) {
			/* Go to closed state. */
			sp->lcp.state = LCP_STATE_CLOSED;
			sp->ipcp.state = IPCP_STATE_CLOSED;
		}

		 /* Initiate renegotiation. */
		wpsppp_clear_timeout (sp);
		wpsppp_set_timeout (sp, 2);
		break;
		
	case LCP_TERM_REQ:
		
		sp->lcp_stats.rx_term_rqst++;
		
		wpsppp_clear_timeout (sp);
		/* Send Terminate-Ack packet. */
	
		sp->lcp_stats.tx_term_ack++;
		wpsppp_cp_send (sp, PPP_LCP, LCP_TERM_ACK, h->ident, 0, 0);
		/* Go to closed state. */
		sp->lcp.state = LCP_STATE_CLOSED;
		sp->ipcp.state = IPCP_STATE_CLOSED;

		/* Added by Nenad Corbic
		 * The Link should go down if LCP_TERM_REQ is
		 * received. */
		if (sp->pp_link_state){ 	
			wpabs_debug_event( "%s: protocol down (rx term req)\n", sp->name);
			sp->pp_link_state=SPPP_LINK_DOWN;
			wpsppp_disconnect_indication(sp, 0);

			/* Added by David Rokhvarg: set authentication bit too */
			if (sp->myauth.proto){
				/* make sure LCP will NOT come up before PAP/CHAP finished */
				sp->pp_flags |= PP_NEEDAUTH;
				sp->myauth.authenticated = 0;
			}
		}
		
		/* Initiate renegotiation. */
		wpsppp_lcp_open (sp);
		break;
	case LCP_TERM_ACK:
		sp->lcp_stats.rx_term_ack++;
		break;
		
	case LCP_CODE_REJ:
		sp->lcp_stats.rx_code_rej++;
		break;
		
	case LCP_PROTO_REJ:
		sp->lcp_stats.rx_proto_rej++;
		/* Ignore for now. */
		break;
		
	case LCP_DISC_REQ:
		sp->lcp_stats.rx_disc_rqst++;
		/* Discard the packet. */
		break;
		
	case LCP_ECHO_REQ:
		if (sp->lcp.state != LCP_STATE_OPENED){
			break;
		}
		
		if (len < 8) {
			sp->lcp_stats.rx_unknown++;
			if (sp->pp_flags & PP_DEBUG) {
				wpabs_debug_event( "%s: invalid lcp echo request packet length: %d bytes\n",
					sp->name, len);
			}
			break;
		}

		sp->lcp_stats.rx_echo_rqst++;
		
		if (wpabs_ntohl (*(int*)(h+1)) == sp->lcp.magic) {
			/* Line loopback mode detected. */
			wpabs_debug_event( "%s: loopback\n", sp->name);
			sp->pp_link_state=SPPP_LINK_DOWN;
			wpsppp_disconnect_indication(sp, 0);

			/* Shut down the PPP link. */
			sp->lcp.state = LCP_STATE_CLOSED;
			sp->ipcp.state = IPCP_STATE_CLOSED;
			wpsppp_clear_timeout (sp);
			/* Initiate negotiation. */
			wpsppp_lcp_open (sp);
			break;
		}
		*(int*)(h+1) = wpabs_htonl (sp->lcp.magic);
		
		sp->lcp_stats.tx_echo_reply++;
		wpsppp_cp_send (sp, PPP_LCP, LCP_ECHO_REPLY, h->ident, len-4, h+1);
		break;
		
	case LCP_ECHO_REPLY:
		
		if (h->ident != sp->lcp.echoid){
			sp->lcp_stats.rx_unknown++;
			break;
		}
		
		if (len < 8){
			sp->lcp_stats.rx_unknown++;
			if (sp->pp_flags & PP_DEBUG) {
				wpabs_debug_event( "%s: invalid lcp echo reply packet length: %d bytes\n",
					sp->name, len);
			}

			break;
		}
		
		sp->lcp_stats.rx_echo_reply++;
		
		if (wpabs_ntohl (*(int*)(h+1)) != sp->lcp.magic){
			sp->pp_alivecnt = 0;
		}

		break;
	}/* switch (h->type) */
}




/**
 *	wp_sppp_input -	receive and process a WAN PPP frame
 *	@skb:	The buffer to process
 *	@dev:	The device it arrived on
 *
 *	This can be called directly by cards that do not have
 *	timing constraints but is normally called from the network layer
 *	after interrupt servicing to process frames queued via netif_rx().
 *
 *	We process the options in the card. If the frame is destined for
 *	the protocol stacks then it requeues the frame for the upper level
 *	protocol. If it is a control from it is processed and discarded
 *	here.
 */


void wp_sppp_input (void *sppp_ptr, void *skb)
{
	struct ppp_header *h;
	struct sppp *sp = (struct sppp *)sppp_ptr;

	wpabs_debug_test("%s: Rx Data len = %i\n",
			__FUNCTION__,wpabs_skb_len(skb));
	
	/* Count received bytes, add FCS and one flag */
	sp->ibytes+= wpabs_skb_len(skb) + 3;
	sp->ipkts++;

	if (wpabs_skb_len(skb) <= PPP_HEADER_LEN) {
		/* Too small packet, drop it. */
		sp->packet_stats.rx_bad_header++;
		if (sp->pp_flags & PP_DEBUG) {
			wpabs_debug_event( "%s: input packet is too small, %d bytes\n",
				sp->name, wpabs_skb_len(skb));
		}
drop:
		wpabs_skb_free(skb);
		return;
	}

	/* Get PPP header. */
	h = (struct ppp_header *)wpabs_skb_data(skb);
	wpabs_skb_pull(skb,sizeof(struct ppp_header));

	switch (h->address) {
	default:        /* Invalid PPP packet. */
		sp->packet_stats.rx_prot_unknwn++;
invalid:
		if (sp->pp_flags & PP_DEBUG) {
			wpabs_debug_event( "%s: invalid input packet <0x%x 0x%x 0x%x>\n",
				sp->name,
				h->address, h->control, wpabs_ntohs (h->protocol));
		}

		goto drop;

	case PPP_ALLSTATIONS:
		
		if (h->control != PPP_UI){
			sp->packet_stats.rx_bad_header++;
			sp->chdlc_stats.Rx_frm_incomp_CHDLC_hdr_count++;
			goto invalid;
		}
	
		if (sp->pp_flags & PP_CISCO) {
			if (sp->pp_flags & PP_DEBUG) {
				wpabs_debug_event( "%s: PPP packet in Cisco mode <0x%x 0x%x 0x%x>\n",
					sp->name,
					h->address, h->control, wpabs_ntohs (h->protocol));
			}

			sp->packet_stats.rx_bad_header++;
			sp->chdlc_stats.Rx_frm_incomp_CHDLC_hdr_count++;
			goto drop;
		}

		switch (wpabs_ntohs (h->protocol)) {
		
		default:
			if (sp->lcp.state == LCP_STATE_OPENED){
				sp->lcp_stats.tx_proto_rej++;
				wpsppp_cp_send (sp, PPP_LCP, LCP_PROTO_REJ,
					++sp->pp_seq[IDX_LCP], wpabs_skb_len(skb) + 2,
					&h->protocol);
			}
			if (sp->pp_flags & PP_DEBUG) {
				wpabs_debug_event( "%s: invalid input protocol <0x%x 0x%x 0x%x>\n",
					sp->name,
					h->address, h->control, wpabs_ntohs (h->protocol));
			}
			
			sp->packet_stats.rx_prot_unknwn++;
			goto drop;
		case PPP_LCP:
			sp->packet_stats.rx_lcp++;
			wpsppp_lcp_input (sp, skb);
			wpabs_skb_free(skb);
			return;
		case PPP_IPCP:
			sp->packet_stats.rx_ipcp++;
			if (sp->lcp.state == LCP_STATE_OPENED){
				wpsppp_ipcp_input (sp, skb);
			}else{
				wpabs_debug_event( "Rx IPCP pkt dropped: waiting LCP finish.\n");
				if (++sp->ipcp_lcp_finish_errors > 5){
						wpabs_debug_event( "Rx IPCP pkt dropped: restarting LCP.\n");
					    sp->ipcp_lcp_finish_errors=0; 

						sp->lcp.state = LCP_STATE_CLOSED;
						sp->ipcp.state = IPCP_STATE_CLOSED;

						sp->pp_link_state=SPPP_LINK_DOWN;
						wpsppp_disconnect_indication(sp, 0);
								
		   			 	/* The link will be renegotiated after timeout */
						wpsppp_clear_timeout (sp);
						wpsppp_set_timeout (sp, 2);           
				}
			}
			wpabs_skb_free(skb);
			return;
		case PPP_IP:
			sp->packet_stats.rx_ip++;
			if (sp->ipcp.state == IPCP_STATE_OPENED) {
				wpsppp_data_indication(sp,skb, WPLIP_IP);
				return;
			}
			break;

		case PPP_IPV6:
			sp->packet_stats.rx_ip++;
			if (sp->ipcp.state == IPCP_STATE_OPENED) {
				wpsppp_data_indication(sp,skb, WPLIP_IPV6);
				return;
			}
			break;

#ifdef IPX
		case PPP_IPX:
			/* IPX IPXCP not implemented yet */
			sp->packet_stats.rx_ipx++;
			if (sp->lcp.state == LCP_STATE_OPENED) {
				wpsppp_data_indication(sp,skb, WPLIP_IPX);
				return;
			}
			break;
#endif
		case PPP_PAP:
			sp->packet_stats.rx_pap++;
			{
				struct lcp_header * lcph = (struct lcp_header *)wpabs_skb_data(skb);
				int len = wpabs_skb_len(skb);
				unsigned char *data = wpabs_skb_data(skb);
			
				switch (lcph->type) {
				
				case PAP_REQ:
					sp->pap_stats.rx_auth_rqst++;
					//Should not get here if peer. It is valid only for authenticator.
					wpabs_debug_event( "%s: Error : Got PAP_REQ !!\n", sp->name);
					break;
					
				case  PAP_ACK:
					sp->pap_stats.rx_auth_ack++;
					data[len] = '\0';
					//print the message in ack packet
					wpabs_debug_event( "%s: Pap Ack: Remote Message : %s\n", sp->name,
						&data[sizeof(struct lcp_header) + sizeof(unsigned char)]);
				
					sp->pp_flags &= ~PP_NEEDAUTH;
					sp->myauth.authenticated = 1;

					/* 6/12/2006 David Rokhvarg: set LCP to 'open' 
					   only when authentication is successful!!! */
					sp->lcp.state = LCP_STATE_OPENED;
					wpsppp_ipcp_open (sp);

					/* indicate protocol 'up' */
					sp->pp_link_state=SPPP_LINK_UP;
					wpabs_debug_event( "%s: protocol up\n", sp->name);
					wpsppp_connect_indication(sp, 0); 
					break;
					
				case  PAP_NAK:
					sp->pap_stats.rx_auth_nak++;
					data[len] = '\0';
					//print the message in nak packet
					wpabs_debug_event( "%s: Pap Nack: Remote Message : %s\n", sp->name,
						&data[sizeof(struct lcp_header) + sizeof(unsigned char)]);

					//await LCP shutdown by authenticator
					break;

				default:
					wpabs_debug_event( "%s : Unknown PAP packet !!!\n", sp->name);
					break;				
				}
			}

			break;

		case PPP_CHAP:
			sp->packet_stats.rx_chap++;
			wpsppp_chap_input(sp, skb);
			break;		
		}
		break;

	case CISCO_MULTICAST:
	case CISCO_UNICAST:

		/* Don't check the control field here (RFC 1547). */
		if (! (sp->pp_flags & PP_CISCO)) {
			if (sp->pp_flags & PP_DEBUG) {
				wpabs_debug_event( "%s: Cisco packet in PPP mode <0x%x 0x%x 0x%x>\n",
					sp->name,
					h->address, h->control, wpabs_ntohs (h->protocol));
			}

			sp->packet_stats.rx_bad_header++;
			goto drop;
		}

		
		
		switch (wpabs_ntohs(h->protocol)) {
		default:
			if (sp->pp_flags & PP_DEBUG) {
				wpabs_debug_event("%s: Cisco: invalid pkt: prot 0x%X\n",
					sp->name,wpabs_ntohs(h->protocol));
			}
			sp->chdlc_stats.Rx_invalid_CHDLC_type_count++;
			goto invalid;

		case CISCO_DISCOVERY:
		        sp->chdlc_stats.Rx_invalid_CHDLC_type_count++;
			goto invalid; 
			
		case CISCO_KEEPALIVE:
			wpsppp_cisco_input (sp, skb);
			wpabs_skb_free(skb);
			return;

		case CISCO_IP:
			sp->chdlc_stats.Data_frames_Rx_count++;
			sp->chdlc_stats.Data_bytes_Rx_count+=wpabs_skb_len(skb);
			wpsppp_data_indication(sp,skb,WPLIP_IP);
			return;

		case CISCO_IPV6:
			sp->chdlc_stats.Data_frames_Rx_count++;
			sp->chdlc_stats.Data_bytes_Rx_count+=wpabs_skb_len(skb);
			wpsppp_data_indication(sp,skb,WPLIP_IPV6);
			return;


		}
		break;
	}
	wpabs_skb_free(skb);
}


