
#include "wanpipe_sppp.h"


/*
 * Timeout routine activations.
 */

void wpsppp_set_timeout(wp_sppp_t *p,int s) 
{
	if (! (p->pp_flags & PP_TIMO)) 
	{
		p->pp_timer=wpabs_get_systemticks();
		p->cfg.pp_timer=s*wpabs_get_hz();
		p->pp_flags |= PP_TIMO;
		DBG_CHDLC("%s(): line: %d: Set Timeout %d\n",
			       	__FUNCTION__, __LINE__,s);
	} else {
                DBG_CHDLC("%s(): line: %d: Timeout already Set\n",
			       	__FUNCTION__, __LINE__);
	}
}

void wpsppp_clear_timeout(wp_sppp_t *p)
{
	if (p->pp_flags & PP_TIMO) 
	{
		p->pp_timer=0;
		p->cfg.pp_timer=10*wpabs_get_hz();
		p->pp_flags &= ~PP_TIMO; 
		DBG_CHDLC("%s(): line: %d: Clear Timeout\n",
			       	__FUNCTION__, __LINE__);  
	} else {
        	DBG_CHDLC("%s(): line: %d: Timeout already cleared\n",
			       	__FUNCTION__, __LINE__);  
	}
}

/*
 * Timeout routines for authentication protocols
 */

void wpsppp_auth_timeout(void * function, wp_sppp_t *p, unsigned int seconds) 
{
	/*do nothing for now!!*/
	return;
	
#if 0
	/*
	** wpabs_debug_event( "auth_timeout() : seconds : %u,  p->pp_auth_flags : 0x%X\n",
	** 	seconds, p->pp_auth_flags); */

	/* only one of them can be waiting to timeout at a time */
	if (! (p->pp_auth_flags & PP_TIMO1)) 
	{
		wpabs_debug_event( "starting pp_auth_timer...\n");
		init_timer(&p->pp_auth_timer);
		p->pp_auth_timer.function = function;
		p->pp_auth_timer.expires  = wpabs_get_systemticks() + seconds * HZ;
		p->pp_auth_timer.data = (unsigned long)p;
		p->pp_auth_flags |= PP_TIMO1;

		
		add_timer(&p->pp_auth_timer);
	}
#endif
}

void wpsppp_auth_clear_timeout(wp_sppp_t *p)
{
	/* do nothing for now!! */
	return;

	/* wpabs_debug_event( "auth_clear_timeout()\n"); */

	if (p->pp_auth_flags & PP_TIMO1) 
	{
		wpabs_debug_event("deleting pp_auth_timer timer...\n");
		p->pp_auth_timer=0;
		p->pp_auth_flags &= ~PP_TIMO1; 
	}
}

 
const char *
wpsppp_auth_type_name(unsigned short proto, unsigned char type)
{
	static char buf[12]="auth/Unknown";
	switch (proto) {
	case PPP_CHAP:
		switch (type) {
		case CHAP_CHALLENGE:	return "challenge";
		case CHAP_RESPONSE:	return "response";
		case CHAP_SUCCESS:	return "success";
		case CHAP_FAILURE:	return "failure";
		}
	case PPP_PAP:
		switch (type) {
		case PAP_REQ:		return "req";
		case PAP_ACK:		return "ack";
		case PAP_NAK:		return "nak";
		}
	}

#if 0
	wpabs_snprintf (buf, sizeof(buf), "auth/0x%x", type);
#endif
	return buf;
}

void
wpsppp_print_string(const char *p, unsigned short len)
{
	unsigned char c;

	while (len-- > 0) {
		c = *p++;
		/*
		 * Print only ASCII chars directly.  RFC 1994 recommends
		 * using only them, but we don't rely on it.  */
		if (c < ' ' || c > '~')
			wpabs_debug_event( "\\x%x", c);
		else
			wpabs_debug_event( "%c", c);
	}
}


/*
 * Change the state of a control protocol in the state automaton.
 * Takes care of starting/stopping the restart timer.
 */
void
wpsppp_cp_change_state(const struct cp *cp, wp_sppp_t *sp, int newstate)
{
	sp->state[cp->protoidx] = newstate;

	UNTIMEOUT((void *)sp);
	
	switch (newstate) {
	case STATE_INITIAL:
	case STATE_STARTING:
	case STATE_CLOSED:
	case STATE_STOPPED:
	case STATE_OPENED:
		break;
	case STATE_CLOSING:
	case STATE_STOPPING:
	case STATE_REQ_SENT:
	case STATE_ACK_RCVD:
	case STATE_ACK_SENT:
		
		TIMEOUT(cp->TO, (void *)sp, sp->lcp.timeout);
		break;
	}
}


/* a dummy, used to drop uninteresting events */
void
wpsppp_null(wp_sppp_t *unused)
{
	wpabs_debug_event( "sppp_null()\n");
	/* UNTIMEOUT( unused ); */
	/* do just nothing */
}

void
wpsppp_phase_network(wp_sppp_t *sp)
{
	/* int i; */
	/* unsigned int mask; */

	sp->pp_phase = PHASE_NETWORK;

#if 0
	wpabs_debug_event( SPP_FMT "phase %s\n", SPP_ARGS(sp),
		sppp_phase_name(sp->pp_phase));
#endif

	/* Notify NCPs now. */
	/* ???????? */
	/*
	for (i = 0; i < IDX_COUNT; i++)
		if ((cps[i])->flags & CP_NCP)
			(cps[i])->Open(sp);
	*/
	/* Send Up events to all NCPs. */
	/* ???????? */
	/*
	for (i = 0, mask = 1; i < IDX_COUNT; i++, mask <<= 1)
		if ((sp->lcp.protos & mask) && ((cps[i])->flags & CP_NCP))
			(cps[i])->Up(sp);
	*/
	/* if no NCP is starting, all this was in vain, close down */
	/* sppp_lcp_check_and_close(sp); */
	
	sp->ipcp.state = IPCP_STATE_CLOSED;
	/* Initiate renegotiation. */
	wpsppp_ipcp_open (sp);
}

/*
 * Check the open NCPs, return true if at least one NCP is open.
 */
#if 0
int
wpsppp_ncp_check(wp_sppp_t *sp)
{
	int i, mask;

	for (i = 0, mask = 1; i < IDX_COUNT; i++, mask <<= 1)
		if ((sp->lcp.protos & mask) && (cps[i])->flags & CP_NCP)
			return 1;
	return 0;
	
	
}
#endif

/*
 * Re-check the open NCPs and see if we should terminate the link.
 * Called by the NCPs during their tlf action handling.
 */
#if 0
void
wpsppp_lcp_check_and_close(wp_sppp_t *sp)
{

	if (sp->pp_phase < PHASE_NETWORK)
		/* don't bother, we are already going down */
		return;

	if (sppp_ncp_check(sp))
		return;

	lcp.Close(sp);
}
#endif

const char *
wpsppp_state_name(int state)
{
	switch (state) {
	case STATE_INITIAL:	return "initial";
	case STATE_STARTING:	return "starting";
	case STATE_CLOSED:	return "closed";
	case STATE_STOPPED:	return "stopped";
	case STATE_CLOSING:	return "closing";
	case STATE_STOPPING:	return "stopping";
	case STATE_REQ_SENT:	return "req-sent";
	case STATE_ACK_RCVD:	return "ack-rcvd";
	case STATE_ACK_SENT:	return "ack-sent";
	case STATE_OPENED:	return "opened";
	}
	return "illegal";
}

const char *
wpsppp_phase_name(enum ppp_phase phase)
{
	switch (phase) {
	case PHASE_DEAD:	return "dead";
	case PHASE_ESTABLISH:	return "establish";
	case PHASE_TERMINATE:	return "terminate";
	case PHASE_AUTHENTICATE: return "authenticate";
	case PHASE_NETWORK:	return "network";
	}
	return "illegal";
}

/************************************
 * Send keepalive packets, every 10 seconds.
 */

void wpsppp_keepalive (wp_sppp_t *sp)
{
	/* Keepalive mode disabled or channel down? */
	if (! (sp->pp_flags & PP_KEEPALIVE)){
		return;
	}

	/* No keepalive in PPP mode if LCP not opened yet. */
	if (! (sp->pp_flags & PP_CISCO) &&
	    sp->lcp.state != LCP_STATE_OPENED){
		return;
	}

	if (sp->pp_alivecnt == sp->sppp_max_keepalive_count) {
		/* No keepalive packets got.  Stop the interface. */
		wpabs_debug_event( "%s: protocol down (keepalive timeout)\n", sp->name);
		sp->pp_link_state=SPPP_LINK_DOWN;
		wpsppp_disconnect_indication(sp, 0);
		if (! (sp->pp_flags & PP_CISCO)) {
			/* Shut down the PPP link. */
			sp->lcp.magic = wpabs_get_systemticks();
			sp->lcp.state = LCP_STATE_CLOSED;
			sp->ipcp.state = IPCP_STATE_CLOSED;
			wpsppp_clear_timeout (sp);
			/* Initiate negotiation. */
			wpsppp_lcp_open (sp);
		}
	}
	if (sp->pp_alivecnt <= sp->sppp_max_keepalive_count){
		++sp->pp_alivecnt;
	}

	if (sp->pp_flags & PP_CISCO){
		wpsppp_cisco_send (sp, CISCO_KEEPALIVE_REQ, ++sp->pp_seq[IDX_LCP],
			sp->pp_rseq[IDX_LCP]);
		sp->chdlc_stats.CHDLC_SLARP_KPALV_Tx_count++;

	}else if (sp->lcp.state == LCP_STATE_OPENED) {
		long nmagic = wpabs_htonl (sp->lcp.magic);
		sp->lcp.echoid = ++sp->pp_seq[IDX_LCP];
		wpsppp_cp_send (sp, PPP_LCP, LCP_ECHO_REQ,
			sp->lcp.echoid, 4, &nmagic);
	}
}



int
wpsppp_strnlen(unsigned char *p, int max)
{
	int len;

	for (len = 0; len < max && *p; ++p)
		++len;
	return len;
}



/*
 * Send PPP LCP packet.
 */
void wpsppp_cp_send (wp_sppp_t *sp, unsigned short proto, unsigned char type,
	unsigned char ident, unsigned short len, void *data)
{
	struct ppp_header *h;
	struct lcp_header *lh;
	void *skb;
	int err;

	skb=wpabs_skb_alloc(PPP_HEADER_LEN+LCP_HEADER_LEN+len);
	if (skb==NULL)
		return;

	h = (struct ppp_header *)wpabs_skb_put(skb, sizeof(struct ppp_header));
	h->address = PPP_ALLSTATIONS;        /* broadcast address */
	h->control = PPP_UI;                 /* Unnumbered Info */
	h->protocol = wpabs_htons (proto);         /* Link Control Protocol */

	lh = (struct lcp_header *)wpabs_skb_put(skb, sizeof(struct lcp_header));
	lh->type = type;
	lh->ident = ident;
	lh->len = wpabs_htons (LCP_HEADER_LEN + len);

	if (len)
		wpabs_memcpy(wpabs_skb_put(skb,len),data, len);

	if (sp->pp_flags & PP_DEBUG) {

		switch (proto){

		case PPP_PAP:
		case PPP_CHAP:
			wpabs_debug_event( "%s: %s output <%s id=0x%x len=%d",
			sp->name,
		    	(proto == PPP_PAP ? "pap" : "chap"),
		    	wpsppp_auth_type_name(proto, type),
		    	lh->ident, len);
			break;
			
		default:
			wpabs_debug_event( "%s: %s output <%s id=%xh len=%xh",
				sp->name, 
				proto==PPP_LCP ? "lcp" : "ipcp",
				proto==PPP_LCP ? wpsppp_lcp_type_name (lh->type) :
				wpsppp_ipcp_type_name (lh->type), lh->ident,
				wpabs_ntohs (lh->len));
			
			break;
		}
		
		if (len)
			wpsppp_print_bytes ((unsigned char*) (lh+1), len);
		wpabs_debug_event (">\n");
	}
	sp->obytes += wpabs_skb_len(skb);

	err=wpsppp_data_transmit(sp,skb);
	if (err) {
     	wpabs_skb_free(skb);
	}
}

/*
 * Send Cisco keepalive packet.
 */

void wpsppp_cisco_send (wp_sppp_t *sp, unsigned int type, unsigned int par1, unsigned int par2)
{
	struct ppp_header *h;
	struct cisco_packet *ch;
	void *skb;
	unsigned long t = wpabs_get_systemticks() * 1000/wpabs_get_hz();
	int err;

	skb=wpabs_skb_alloc(PPP_HEADER_LEN+CISCO_PACKET_LEN);
	if(skb==NULL)
		return;
		
	h = (struct ppp_header *)wpabs_skb_put (skb, sizeof(struct ppp_header));
	h->address = CISCO_MULTICAST;
	h->control = 0;
	h->protocol = wpabs_htons (CISCO_KEEPALIVE);
	
	ch = (struct cisco_packet*)wpabs_skb_put(skb, CISCO_PACKET_LEN);
	ch->type = (unsigned int)wpabs_htonl ((unsigned int)type);
	ch->par1 = (unsigned int)wpabs_htonl ((unsigned int)par1);
	ch->par2 = (unsigned int)wpabs_htonl ((unsigned int)par2);

	ch->rel = -1;
	ch->time0 = wpabs_htons ((unsigned short) (t >> 16));
	ch->time1 = wpabs_htons ((unsigned short) t);

	if (sp->pp_flags & PP_DEBUG)
		wpabs_debug_event( "%s: cisco output: <%xh %xh %xh %xh %xh-%xh>\n",
			sp->name,  wpabs_ntohl (ch->type), ch->par1,
			ch->par2, ch->rel, ch->time0, ch->time1);
	sp->obytes += wpabs_skb_len(skb);

	err=wpsppp_data_transmit(sp,skb);
	if (err) {
     	wpabs_skb_free(skb);
	}
}


/**
 *	wp_sppp_reopen - notify of physical link loss
 *	@dev: Device that lost the link
 *
 *	This function informs the synchronous protocol code that
 *	the underlying link died (for example a carrier drop on X.21)
 *
 *	We increment the magic numbers to ensure that if the other end
 *	failed to notice we will correctly start a new session. It happens
 *	do to the nature of telco circuits is that you can lose carrier on
 *	one endonly.
 *
 *	Having done this we go back to negotiating. This function may
 *	be called from an interrupt context.
 */
 
int wp_sppp_reopen (wp_sppp_t *sp)
{
	wp_sppp_close(sp);
	if (!(sp->pp_flags & PP_CISCO)){
		sp->lcp.magic = wpabs_get_systemticks();
		++sp->pp_seq[IDX_LCP];
		sp->lcp.state = LCP_STATE_CLOSED;
		sp->ipcp.state = IPCP_STATE_CLOSED;
		/* Give it a moment for the line to settle then go */
		wpsppp_set_timeout (sp, 1);
	} 
	sp->pp_link_state=SPPP_LINK_DOWN;
	return 0;
}


/**
 *	wp_sppp_attach - attach synchronous PPP/HDLC to a device
 *	@pd:	PPP device to initialise
 *
 *	This initialises the PPP/HDLC support on an interface. At the
 *	time of calling the dev element must point to the network device
 *	that this interface is attached to. The interface should not yet
 *	be registered. 
 */
 
void wp_sppp_attach(wp_sppp_t *sp)
{
	FUNC_BEGIN();
	
	DBG_CHDLC("2: sp->cfg.sppp_keepalive_timer: %d\n", sp->cfg.sppp_keepalive_timer);
	DBG_CHDLC("   sp->cfg.keepalive_err_margin: %d\n", sp->cfg.keepalive_err_margin);
	/* store original configuration, BEFORE any changes applied */
	wpabs_memcpy(&sp->original_cfg, &sp->cfg, sizeof(wan_sppp_if_conf_t));
	
	if(sp->pp_flags & PP_CISCO){
		DBG_CHDLC("running in CHDLC mode\n");
		/* for CHDLC the keepalive timer is in the 'conf' file, for PPP it's not */
		sp->cfg.sppp_keepalive_timer = sp->cfg.sppp_keepalive_timer * wpabs_get_hz();
		if(sp->cfg.sppp_keepalive_timer){
			sp->sppp_keepalive_timer = wpabs_get_systemticks();
		}else{
			/* this is where the keepalve logic is actually disabled */
			sp->sppp_keepalive_timer = 0;
		}
		sp->sppp_max_keepalive_count = sp->cfg.keepalive_err_margin;
	}else{
		DBG_CHDLC("running in PPP mode\n");
		/* for PPP keepalives (LCP Echo) timers are not configurable */
		sp->cfg.sppp_keepalive_timer = 5*wpabs_get_hz();
		sp->sppp_keepalive_timer = wpabs_get_systemticks();
		sp->sppp_max_keepalive_count = 10;
	}
	
	sp->cfg.pp_timer=5*wpabs_get_hz();

	sp->pp_loopcnt = 0;
	sp->pp_alivecnt = 0;

	sp->pp_seq[IDX_LCP] = 0;
	sp->pp_rseq[IDX_LCP] = 0;

#if 0
	debug=PP_DEBUG;
#endif
	
	if (sp->cfg.pap){
		sp->myauth.proto = PPP_PAP;
	}else if (sp->cfg.chap){
		sp->myauth.proto = PPP_CHAP;
	}else{
		sp->myauth.proto = 0;
	}

	if (sp->myauth.proto){
		/* make sure LCP will NOT come up before PAP/CHAP finished */
		sp->pp_flags |= PP_NEEDAUTH;
		wpabs_memcpy(sp->myauth.name, sp->cfg.userid, AUTHNAMELEN);
		wpabs_memcpy(sp->myauth.secret, sp->cfg.passwd, AUTHNAMELEN);
	}

	wpsppp_chap_init(sp);

	sp->pp_flags |= debug;
	
	sp->pp_auth_flags = 0;
	sp->lcp.magic = 0;
	sp->lcp.state = LCP_STATE_CLOSED;
	sp->ipcp.state = IPCP_STATE_CLOSED;

	sp->dynamic_ip = sp->cfg.dynamic_ip;

	if (wpabs_strlen(sp->name) > WAN_IFNAME_SZ || 
	    wpabs_strlen(sp->name) == 0){
		wpabs_debug_event("SPPP: Warning: hwdevname not initialized!\n");
	}

	sp->task_working=0;
	sp->local_ip=0;
	sp->remote_ip=0;
	sp->task_working=0;


	wpabs_debug_event("%s: Sync %s Configuration\n",
			sp->name, sp->pp_flags&PP_CISCO?"CISCO":"PPP");


	if (!(sp->pp_flags&PP_CISCO)){
		
		wpabs_debug_event("%s:    IP Mode           :%s\n",
				sp->name, sp->dynamic_ip == WANOPT_PPP_PEER ?"Dynamic": 
				          sp->dynamic_ip == WANOPT_PPP_HOST ?"Static Host" : "Static");

		wpabs_debug_event("%s:    PPP Timeout Timer :%d\n",
				sp->name, sp->cfg.pp_timer/wpabs_get_hz());

		wpabs_debug_event("%s:    Authentication    :%s\n",
				sp->name, sp->myauth.proto == PPP_PAP?"PAP":
					  sp->myauth.proto == PPP_CHAP?"CHAP":
					  "None");

		wpabs_debug_event("%s:    Auth Timer        :%d\n",
				sp->name, sp->cfg.pp_auth_timer/wpabs_get_hz());

		wpabs_debug_event("%s:    Magic 			:%s\n",
				sp->name, sp->cfg.disable_magic?"Off":"On");
	}else{

		wpabs_debug_event("%s:    Keep Alive Timer      :%d\n",
				sp->name, sp->cfg.sppp_keepalive_timer/wpabs_get_hz());

		wpabs_debug_event("%s:    Keep Alive Err Margin :%d\n",
				sp->name, sp->sppp_max_keepalive_count);
	}

	FUNC_END();
}


/**
 *	wp_sppp_detach - release PPP resources from a device
 *	@dev:	Network device to release
 *
 *	Stop and free up any PPP/HDLC resources used by this
 *	interface. This must be called before the device is
 *	freed.
 */
 
void wp_sppp_detach (void *sp_ptr)
{
	wp_sppp_t *sp = (wp_sppp_t *)sp_ptr;
	
	/* Stop keepalive handler. */
	wpsppp_clear_timeout (sp);
	wpsppp_auth_clear_timeout(sp);
}



/*
 * Analyze the LCP Configure-Request options list
 * for the presence of unknown options.
 * If the request contains unknown options, build and
 * send Configure-reject packet, containing only unknown options.
 */
int
wpsppp_lcp_conf_parse_options (wp_sppp_t *sp, struct lcp_header *h,
	int len, unsigned int *magic)
{
	unsigned char *buf, *r, *p;
	int rlen = 0;

	unsigned char * packet_length, * chap_algorithm;
	unsigned short authproto;

	len -= 4; /* sizeof(struct lcp_header); */
	buf = r = wpabs_malloc (len);
	if (! buf)
		return (0);

	p = (void*) (h+1);
	
	/* p[0] is type */
	/* p[1] is length */
	
	for (rlen=0; len>1 && p[1]; len-=p[1], p+=p[1]) {
		switch (*p) {
		case LCP_OPT_MAGIC:
			/* Magic number -- extract. */
			if (len >= 6 && p[1] == 6) {
				*magic = (unsigned int)p[2] << 24 |
					(unsigned int)p[3] << 16 | p[4] << 8 | p[5];
				continue;
			}
			break;
		case LCP_OPT_ASYNC_MAP:
			/* Async control character map -- check to be zero. */
			if (len >= 6 && p[1] == 6 && ! p[2] && ! p[3] &&
			    ! p[4] && ! p[5])
				continue;
			break;
		case LCP_OPT_MRU:
			/* Maximum receive unit -- always OK. */
			continue;

		case LCP_OPT_AUTH_PROTO:
			/* request to authenticate */
			packet_length = p + 1;
			
			if (*packet_length < 4) {
				wpabs_debug_event( "%s: LCP_OPT_AUTH_PROTO packet is too short (%d)!\n",
					sp->name, *packet_length);
				break;
			}
			
			/* got request to authenticate */
			if (sp->myauth.proto == 0) {
				/* we are not configured to do auth */
				wpabs_debug_event(
			       		"%s: Got authentication request, but not configured to authenticate.\n", 
					sp->name);
				break;
			}
		
			authproto  = wpabs_ntohs(*(unsigned short *)(p + 2));
			/* wpabs_debug_event( "authentication protocol : 0x%X\n", authproto); */

			if (sp->myauth.proto != authproto) {
				/* not agreed, nak */
				wpabs_debug_event(
				"%s: Authentication protocols mismatch! (mine : %s , his : %s )\n",
						sp->name,
					       	wpsppp_proto_name(sp->myauth.proto),
					       	wpsppp_proto_name(authproto));
				break;
			}

			/*
			 * Remote wants us to authenticate, remember this,
			 * so we authenticate once LCP is complete in both directions
			 * since we don't really have a PHASE_AUTHENTICATE. 
			 */
			DBG_CHDLC("%s(): line: %d\n", __FUNCTION__, __LINE__);

			sp->pp_flags |= PP_NEEDAUTH;

			/* if we configured for PAP -- ack it */
			if(authproto == PPP_PAP){
				DBG_CHDLC("%s(): line: %d\n", __FUNCTION__, __LINE__);
				wpabs_free(buf);
				return PPP_PAP;
			}else if(authproto == PPP_CHAP){
				DBG_CHDLC("%s(): line: %d\n", __FUNCTION__, __LINE__);

				chap_algorithm = p + 4;

				if(*chap_algorithm == CHAP_MD5){

					wpabs_free(buf);
					return 1;
				}else{
					wpabs_debug_event( "Got request for unsupported CHAP algorithm : 0x%X.\
							Only MD5 is supported.\n", *chap_algorithm);
					break;
				}
			}else{
				wpabs_debug_event( "Got request for unsupported authentication protocol : 0x%X\n", 
					authproto);
				break;
			}
				
		default:
			/* Others not supported. */
			wpabs_debug_event( "unknown lcp option : 0x%X\n", *p);
			break;
		}
		/* Add the option to rejected list. */
		wpabs_memcpy(r, p, p[1]);
		r += p[1];
		rlen += p[1];
	}
	if (rlen)
		wpsppp_cp_send (sp, PPP_LCP, LCP_CONF_REJ, h->ident, rlen, buf);

	wpabs_free(buf);
	return (rlen == 0);
}

const char *
wpsppp_proto_name(unsigned short proto)
{
	static char buf[20]="proto/Unknown";
	switch (proto) {
	case PPP_LCP:	return "lcp";
	case PPP_IPCP:	return "ipcp";
	case PPP_PAP:	return "pap";
	case PPP_CHAP:	return "chap";
	/* case PPP_IPV6CP: return "ipv6cp"; */
	}
#if 0
	wpabs_snprintf(buf, sizeof(buf), "proto/0x%x", (unsigned)proto);
#endif
	return buf;
}


void
byteSwap(unsigned int *buf, unsigned words)
{
#ifdef WORDS_BIGENDIAN
	md5byte *p = (md5byte *)buf;

	do {
		*buf++ = (unsigned int)((unsigned)p[3] << 8 | p[2]) << 16 |
			((unsigned)p[1] << 8 | p[0]);
		p += 4;
	} while (--words);
#endif
}



void wpsppp_lcp_open (wp_sppp_t *sp)
{
	char opt[6];

	if (sp->dynamic_ip == WANOPT_PPP_PEER){
		sp->local_ip=0;
		sp->remote_ip=0;
		if (wpsppp_get_ipv4addr(sp,WAN_LOCAL_IP)){
			sp->dynamic_event=1;
		}
	}

	if (sp->cfg.disable_magic == WANOPT_NO) {
	
		if (! sp->lcp.magic)
			sp->lcp.magic = wpabs_get_systemticks();
		opt[0] = LCP_OPT_MAGIC;
		opt[1] = sizeof (opt);
		opt[2] = sp->lcp.magic >> 24;
		opt[3] = sp->lcp.magic >> 16;
		opt[4] = sp->lcp.magic >> 8;
		opt[5] = sp->lcp.magic;
	}

	sp->lcp.confid = ++sp->pp_seq[IDX_LCP];

	sp->lcp_stats.tx_conf_rqst++;

	if (sp->cfg.disable_magic == WANOPT_NO) {
		wpsppp_cp_send (sp, PPP_LCP, LCP_CONF_REQ, sp->lcp.confid,
			sizeof (opt), &opt);
	} else {
		wpsppp_cp_send (sp, PPP_LCP, LCP_CONF_REQ, sp->lcp.confid,
			0, NULL);
	}
	wpsppp_set_timeout (sp, 2);
}

void wpsppp_ipcp_open (wp_sppp_t *sp)
{
        unsigned char data[]={3,6,0,0,0,0};
	unsigned int ip=wpsppp_get_ipv4addr(sp,WAN_LOCAL_IP);
	
	if (sp->dynamic_ip == WANOPT_PPP_PEER){
		sp->ipcp.confid = ++sp->pp_seq[IDX_LCP];
		sp->prot_stats.tx_conf_rqst++;
		wpsppp_cp_send (sp, PPP_IPCP, IPCP_CONF_REQ, sp->ipcp.confid, 6, data);
		wpsppp_set_timeout (sp, 10);

	}else{
		sp->ipcp.confid = ++sp->pp_seq[IDX_LCP];
		sp->prot_stats.tx_conf_rqst++;
		if (sp->dynamic_ip == WANOPT_PPP_HOST && ip) {
			*(unsigned int*)&data[2]=ip;
	       	 	wpsppp_cp_send (sp, PPP_IPCP, IPCP_CONF_REQ, sp->ipcp.confid, 6,data);
		} else {
                	wpsppp_cp_send (sp, PPP_IPCP, IPCP_CONF_REQ, sp->ipcp.confid, 0,0); 
		}
		wpsppp_set_timeout (sp, 10);
	}
}

/*
 * Process PPP control protocol timeouts.
 */
 
void wpsppp_cp_timeout (wp_sppp_t *sp)
{
	sp->pp_flags &= ~PP_TIMO;
	if ((sp->pp_flags & PP_CISCO)) {
		return;
	}

	 DBG_CHDLC("%s(): line: %d  PPP Timeout: LCP=%d IPCP=%d \n",
			 __FUNCTION__, __LINE__,
			 sp->lcp.state,sp->ipcp.state);
	 
	switch (sp->lcp.state) {
	case LCP_STATE_CLOSED:
		/* No ACK for Configure-Request, retry. */
		wpsppp_lcp_open (sp);
		break;
	case LCP_STATE_ACK_RCVD:
		/* ACK got, but no Configure-Request for peer, retry. */
		wpsppp_lcp_open (sp);
		sp->lcp.state = LCP_STATE_CLOSED;
		break;
	case LCP_STATE_ACK_SENT:
		/* ACK sent but no ACK for Configure-Request, retry. */
		wpsppp_lcp_open (sp);
		break;
	case LCP_STATE_OPENED:
		/* LCP is already OK, try IPCP. */
		switch (sp->ipcp.state) {
		case IPCP_STATE_CLOSED:
			/* No ACK for Configure-Request, retry. */
			wpsppp_ipcp_open (sp);
			break;
		case IPCP_STATE_ACK_RCVD:
			/* ACK got, but no Configure-Request for peer, retry. */
			wpsppp_ipcp_open (sp);
			sp->ipcp.state = IPCP_STATE_CLOSED;
			break;
		case IPCP_STATE_ACK_SENT:
			/* ACK sent but no ACK for Configure-Request, retry. */
			wpsppp_ipcp_open (sp);
			break;
		case IPCP_STATE_OPENED:
			/* IPCP is OK. */
			break;
		}
		break;
	}
}

char *wpsppp_lcp_type_name (unsigned char type)
{
	static char buf [8]="Unknown";
	switch (type) {
	case LCP_CONF_REQ:   return ("conf-req");
	case LCP_CONF_ACK:   return ("conf-ack");
	case LCP_CONF_NAK:   return ("conf-nack");
	case LCP_CONF_REJ:   return ("conf-rej");
	case LCP_TERM_REQ:   return ("term-req");
	case LCP_TERM_ACK:   return ("term-ack");
	case LCP_CODE_REJ:   return ("code-rej");
	case LCP_PROTO_REJ:  return ("proto-rej");
	case LCP_ECHO_REQ:   return ("echo-req");
	case LCP_ECHO_REPLY: return ("echo-reply");
	case LCP_DISC_REQ:   return ("discard-req");
	}

#if 0
	wpabs_sprintf (buf, "%xh", type);
#endif	
	return (buf);
}

char *wpsppp_ipcp_type_name (unsigned char type)
{
	static char buf [8]="Unknown";
	switch (type) {
	case IPCP_CONF_REQ:   return ("conf-req");
	case IPCP_CONF_ACK:   return ("conf-ack");
	case IPCP_CONF_NAK:   return ("conf-nack");
	case IPCP_CONF_REJ:   return ("conf-rej");
	case IPCP_TERM_REQ:   return ("term-req");
	case IPCP_TERM_ACK:   return ("term-ack");
	case IPCP_CODE_REJ:   return ("code-rej");
	}
#if 0
	wpabs_sprintf (buf, "%xh", type);
#endif
	return (buf);
}

void wpsppp_print_bytes (unsigned char *p, unsigned short len)
{
	wpabs_debug_event (" %x", *p++);
	while (--len > 0)
		wpabs_debug_event ("-%x", *p++);
}


char *decode_lcp_state(int state)
{
	switch (state){

		case LCP_STATE_CLOSED:        	/* LCP state: closed (conf-req sent) */
			return "closed";
		case LCP_STATE_ACK_RCVD:	/* LCP state: conf-ack received */
			return "conf-ack rx";
		case LCP_STATE_ACK_SENT:      	/* LCP state: conf-ack sent */
			return "conf-ack tx";
		case LCP_STATE_OPENED:		/* LCP state: opened */
			return "opened";
	}
	return "invalid";
}

char *decode_ipcp_state(int state)
{
	switch (state){

		case IPCP_STATE_CLOSED:      	/* IPCP state: closed (conf-req sent) */
			return "closed";
		case IPCP_STATE_ACK_RCVD:	/* IPCP state: conf-ack received */
			return "conf-ack rx";
		case IPCP_STATE_ACK_SENT:	/* IPCP state: conf-ack sent */
			return "conf-ack tx";
		case IPCP_STATE_OPENED:	 	/* IPCP state: opened */
			return "opened";
	}
	return "invalid";
}



wp_sppp_t *wpsppp_create_cb(void)
{
	wp_sppp_t *sp;
	
	sp=wpabs_malloc(sizeof(wp_sppp_t));
	if (sp == NULL)
		return NULL;

	wpabs_memset(sp, 0x00, sizeof(wp_sppp_t));

	wp_dev_hold(sp);

	return sp;
}



void wpsppp_free_cb(wp_sppp_t *sp)
{

	wp_dev_put(sp);

	return;
}

void wpsppp_bh (struct sppp *sp)
{
	int err;
	
	if (sp->dynamic_ip != WANOPT_PPP_PEER) {
		return;
	}

	if (sp->ipcp.state == IPCP_STATE_OPENED){

		wpabs_debug_event("%s: Local=%u.%u.%u.%u  Remote IP=%u.%u.%u.%u\n",
					sp->name,
					NIPQUAD(sp->local_ip),
					NIPQUAD(sp->remote_ip));

		err=wpsppp_set_ipv4addr(sp,sp->local_ip,sp->remote_ip,0,0);
		if (err){
			wpabs_debug_event("%s: Failed to set Local/Remote IP: Rc=%i\n",
					sp->name,err);
		}
		
		if (sp->gateway){
			wpsppp_set_gateway(sp);
		}
#if 0
CALL THE USER LEVEL CODE!!!
		wan_run_wanrouter(sp->hwdevname,sp->name,"script");
#endif

	}else{
		sp->local_ip=0;
		sp->remote_ip=0;
		wpabs_debug_event("%s: Local IP Address %u.%u.%u.%u\n",
					sp->name,NIPQUAD(sp->local_ip));

		if (wpsppp_get_ipv4addr(sp,WAN_LOCAL_IP)){
			wpabs_debug_event("%s: Local IP Address Exists Removing\n",
					sp->name);
			wpsppp_set_ipv4addr(sp,sp->local_ip,0,0,0);
		}	
	}
}

